def _log_msg(name: str, size: int) -> str:  # pragma: no cover
    """Return log message for creation of file.

    Args:
        name (str): Name of the created file.
        size (int): Size of the created file.

    Returns:
        str: Log message with information about the created file.
    """
    i = 0
    units = ['B', 'KB', 'MB', 'GB', 'TB']
    while size >= 1000:
        size = round(size / 1000)
        i += 1
    size_str = '%d%s' % (size, units[i])
    return "%s | %s" % (name, size_str)



import numpy as np
import os
import pkgutil
import copy
from . import netpbm
from ._log import _log_msg
import logging


# Create a map from ascii characters to their image representation.
# Here are some scaled-down examples of the mappings in CHAR_TO_IMG.
#
#            0 0 0 0 0 0 0             0 0 1 0 1 0 0
#            0 0 1 0 0 0 0             0 1 1 1 1 1 0
#    ~  ->   0 1 0 1 0 1 0     #  ->   0 0 1 0 1 0 0
#            0 0 0 0 1 0 0             0 1 1 1 1 1 0
#            0 0 0 0 0 0 0             0 0 1 0 1 0 0

file = pkgutil.get_data(__name__, "resources/ascii.pgm").decode().split('\n')
ascii_M = netpbm._parse_ascii_netpbm(file).M
char_images = [np.pad(M,((0,0),(6,6))) for M in np.split(ascii_M, 13, axis=1)]
CHAR_TO_IMG = dict(zip(list(" .,-~:;=!*#$@"), char_images))


class Ascii:
    """An object representing an ASCII image.

    For more information about ASCII, see
    `ASCII <https://wikipedia.org/wiki/ASCII>`_
    """
    def __init__(self, M: np.ndarray):
        """Initialize an ASCII image.

        Args:
            M (np.ndarray): A NumPy array of ASCII characters.
        """
        self.M = M

    def to_txt(self, path: str):
        """Write object to a txt file.

        Args:
            path (str): String file path.
        """
        with open(path, "w") as f:
            lines = self.M.astype(str).tolist()
            f.write('\n'.join([' '.join(line) for line in lines]))
            f.write('\n')
        logging.info(_log_msg(path, os.stat(path).st_size))

    def to_png(self, path: str):
        """Write object to a png file.

        Args:
            path (str): String file path.
        """
        A = self.M
        n,m = A.shape
        M = []
        for i in range(n):
            M.append([CHAR_TO_IMG[A[i,j]] for j in range(m)])
        M = np.block(M)
        image = netpbm.Netpbm(P=2, k=255, M=M)
        image.to_png(path)
        logging.info(_log_msg(path, os.stat(path).st_size))


def netpbm_to_ascii(image: netpbm.Netpbm) -> Ascii:
    """Return an ASCII representation of the given image.

    This function uses a particular style of
    `ASCII art <https://en.wikipedia.org/wiki/ASCII_art>`_
    in which "symbols with various intensities [are used for] creating
    gradients or contrasts."

    Args:
        image (netpbm.Netpbm): Netpbm image.

    Returns:
        Ascii: ASCII representation of image.
    """
    chars = "  -~:;=!*#$@"
    image = copy.copy((image))
    image.set_max_color_value(len(chars)-1)
    M = image.M
    M = np.array([[chars[i] for i in row] for row in M])
    return Ascii(M=M)



"""
dmtools
=======

dmtools (Digital Media Tools) is a Python package providing low-level tools for
working with digital media programmatically. The netpbm module allows one to
read and create Netpbm images. Color space transformations can be done with the
colorspace module. Using ffmpeg, the animation module can export .mp4 videos
formed from a list of images and the sound module can be used to add sound to
these videos as well. Lastly, ASCII art can be produced with the ascii module.
"""

__author__ = 'Henry Robbins'

from . import animation
from . import ascii
from . import colorspace
from . import netpbm
from . import sound
from . import arrange
from . import transform
from .io import read_png, write_png



import numpy as np
from math import floor, ceil, sqrt
from functools import partial
from typing import Callable


def _box_weighting_function(x: float) -> float:
    """Box filter's weighting function.

    For more information about the Box filter, see
    `Box <https://legacy.imagemagick.org/Usage/filter/#box>`_.

    Args:
        x (float): distance to source pixel.

    Returns:
        float: weight on the source pixel.
    """
    return 1 if x <= 0.5 else 0


def _triangle_weighting_function(x: float) -> float:
    """Triangle filter's weighting function.

    For more information about the Triangle filter, see
    `Triangle <https://legacy.imagemagick.org/Usage/filter/#triangle>`_.

    Args:
        x (float): distance to source pixel.

    Returns:
        float: weight on the source pixel.
    """
    return max(1 - x, 0.0)


def _catmull_rom_weighting_function(x: float) -> float:
    """Catmull-Rom filter's weighting function.

    For more information about the Catmull-Rom filter, see
    `Cubic Filters <https://legacy.imagemagick.org/Usage/filter/#cubics>`_.

    Args:
        x (float): distance to source pixel.

    Returns:
        float: weight on the source pixel.
    """
    if x <= 1:
        return (3*x**3 - 5*x**2 + 2) / 2
    elif x <= 2:
        return (-x**3 + 5*x**2 - 8*x + 4) / 2
    else:
        return 0


def _gaussian_weighting_function(x: float,
                                 sigma: float = 0.5,
                                 blur: float = 1.0) -> float:
    """Gaussian blur function.

    For information about Gaussian blur, see
    `Gaussian <https://legacy.imagemagick.org/Usage/filter/#gaussian>`_.

    Args:
        x (float): Distance to source pixel.
        sigma (float): Determines the "neighborhood" of blur. Defaults to 0.5.
        blur (float): Scale sigma by some multiplier. Defaults to 1.0.

    Returns:
        float: weight on the source pixel.
    """
    sigma = sigma * blur
    return (1 / sqrt(2*np.pi*sigma**2))*np.power(np.e, -x**2 / (2*sigma**2))


RESIZE_FILTERS = \
    {'point':    (_box_weighting_function,          0.0),
     'box':      (_box_weighting_function,          0.5),
     'triangle': (_triangle_weighting_function,     1.0),
     'catrom':   (_catmull_rom_weighting_function,  2.0),
     'gaussian': (_gaussian_weighting_function,     2.0)}


def _rescale_axis(image: np.ndarray,
                  axis: int,
                  k: int,
                  filter: str,
                  weighting_function: Callable = None,
                  support: Callable = None,
                  **kwargs) -> np.ndarray:
    # set the weighting function and support
    if weighting_function is not None:
        if support is None:
            raise ValueError('Weighting function provided without support.')
        f = weighting_function
        support = support
    else:
        f, support = RESIZE_FILTERS[filter]

    # scale support if blur keyword argument is passed
    if 'blur' in kwargs:
        support = support * kwargs['blur']

    if k > 1:
        support = support * k

    if axis == 1:
        image = np.swapaxes(image,0,1)

    n, *_ = image.shape
    new_shape = list(image.shape)
    new_shape[0] = int(new_shape[0] * k)
    rescaled_image = np.zeros(new_shape)
    for i in range(new_shape[0]):
        # get range of rows in the support
        bisect = i + 0.5
        a = max((bisect - support) / k, 0.0)
        b = min((bisect + support) / k, n)
        if (b-a < 1):
            # fall back to nearest neighbor heuristic
            if ceil(a) - a > ((b - a) / 2.0):
                a = floor(a)
            else:
                a = ceil(a)
            b = a + 1
        a = round(a)
        b = round(b)
        row = image[a:b,:]

        def x(i):
            """Return distance to source pixel."""
            return abs((i+0.5) - (bisect / k))

        # use weighting function to weight rows
        if k <= 1:
            weights = [f(x(i) * k, **kwargs) for i in range(a,b)]
        else:
            weights = [f(x(i), **kwargs) for i in range(a,b)]
        row = np.average(row, axis=0, weights=weights)

        # set row of rescaled image
        rescaled_image[i,:] = row

    if axis == 1:
        rescaled_image = np.swapaxes(rescaled_image,0,1)

    return rescaled_image


def rescale(image: np.ndarray,
            k: int,
            filter: str = 'point',
            weighting_function: Callable = None,
            support: Callable = None,
            **kwargs) -> np.ndarray:
    """Rescale the image by the given scaling factor.

    This image rescale implentation is largley based off of the `ImageMagick`_
    impmenetation. The following filters are built-in:

    - `Point Filter`_ ("point"): Nearest-neighbor heuristic.
    - `Box Filter`_ ("box"): Average of neighboring pixels.
    - `Triangle Filter`_ ("triangle"): Linear decrease in pixel weight.
    - `Catmull-Rom Filter`_ ("catrom"): Produces a sharper edge.
    - `Gaussian Filter`_ ("gaussian"): Blurs image. Useful as low pass filter.

    Additionally, advanced users can specify a custom filter by providing a
    weighting function and a support.

    .. _ImageMagick: https://imagemagick.org/script/index.php
    .. _Point Filter: https://legacy.imagemagick.org/Usage/filter/#point
    .. _Box Filter: https://legacy.imagemagick.org/Usage/filter/#box
    .. _Triangle Filter: https://legacy.imagemagick.org/Usage/filter/#triangle
    .. _Catmull-Rom Filter: https://legacy.imagemagick.org/Usage/filter/#cubics
    .. _Gaussian Filter: https://legacy.imagemagick.org/Usage/filter/#gaussian

    Args:
        image (np.ndarray): Image to rescale.
        k (int): Scaling factor.
        filter (str): {point, box, triangle, catrom, gaussian}.
        weighting_function (Callable): Weighting function to use.
        support (float): Support of the provided weighting function.

    Returns:
        np.ndarray: Rescaled image.
    """
    rescaled_image = _rescale_axis(image=image, axis=0, k=k, filter=filter,
                                   weighting_function=weighting_function,
                                   support=support, **kwargs)
    rescaled_image = _rescale_axis(image=rescaled_image, axis=1, k=k,
                                   filter=filter,
                                   weighting_function=weighting_function,
                                   support=support, **kwargs)
    return rescaled_image


def blur(image: np.ndarray, sigma: float, radius: float = 0) -> np.ndarray:
    """Blur the image.

    This image blur implentation is largley based off of the `ImageMagick`_
    impmenetation. It uses a `Gaussian Filter`_ with parameter ``sigma`` and
    a support of ``radius`` to blur the image.

    .. _ImageMagick: https://imagemagick.org/script/index.php
    .. _Gaussian Filter: https://legacy.imagemagick.org/Usage/filter/#gaussian

    Args:
        image (np.ndarray): Image to be blurred.
        sigma (float): "Neighborhood" of the blur. A larger value is blurrier.
        radius (float): Limit of the blur. Defaults to 4 x sigma.

    Returns:
        np.ndarray: Blurred image.
    """
    if radius == 0:
        radius = 4 * sigma
    f = partial(_gaussian_weighting_function, sigma=sigma)
    return rescale(image, k=1, weighting_function=f, support=radius)


def clip(image: np.ndarray, k:int = 255) -> np.ndarray:
    """Clip the image so the maximum gray/color value is k.

    Every value less than 0 is mapped to 0 and every value more than k is
    mapped to k. Values in [0,k] are untouched.

    Args:
        image (np.ndarray): Image to clip.
        k (int): Maximum gray/color value. Defaults to 255.

    Returns:
        np.ndarray: Clipped image.
    """
    return np.clip(image, 0, k)


def normalize(image: np.ndarray, k:int = 255) -> np.ndarray:
    """Normalize the image so the maximum gray/color value is k.

    Normalize the range of values in the image to [0,k]. If applied to a
    three channel image, normalizes each channel by the same amount.

    Args:
        image (np.ndarray): Image to normalize.
        k (int): Maximum gray/color value. Defaults to 255.

    Returns:
        np.ndarray: Normalized image.
    """
    if np.max(image) == np.min(image):
        # every value in the image is the same--fall back to clip
        return clip(image, k)
    image = image - np.min(image)
    return image * (k / (np.max(image)))


def wraparound(image: np.ndarray, k:int = 255) -> np.ndarray:
    """Wraparound the image so the maximum gray/color value is k.

    Each value x is mapped to x mod k+1 such that values outside of [0,k]
    wraparound until they fall in the desired range.

    Args:
        image (np.ndarray): Image to wraparound
        k (int): Maximum gray/color value. Defaults to 255.

    Returns:
        np.ndarray: Wraparound image.
    """
    return np.mod(image, k+1)



import os
import pytest
import numpy as np
from dmtools.transform import rescale, blur, clip, normalize, wraparound
from dmtools.colorspace import gray_to_RGB
from dmtools.io import read_png

RESOURCES_PATH = os.path.join(os.path.dirname(__file__), 'resources')

# These tests are derived from ImageMagick example images which can be found in
# https://legacy.imagemagick.org/Usage/filter/
#
# Links test imagery used for each filter
#
# Point filter: https://legacy.imagemagick.org/Usage/filter/#point
# Box filter: https://legacy.imagemagick.org/Usage/filter/#box
# Triangle filter: https://legacy.imagemagick.org/Usage/filter/#triangle
# Gaussian filter: https://legacy.imagemagick.org/Usage/filter/#gaussian


@pytest.mark.parametrize("image,filter,k,new_name",[
    ('checks_10', 'point', 0.9, 'point_0.9'),
    ('checks_10', 'point', 0.8, 'point_0.8'),
    ('checks_10', 'point', 0.7, 'point_0.7'),
    ('checks_10', 'point', 0.6, 'point_0.6'),
    ('checks_10', 'point', 0.5, 'point_0.5'),
    ('checks_10', 'box', 0.9, 'box_0.9'),
    # ('checks_10', 'box', 0.8, 'box_0.8'), strangely differs from ImageMagick
    ('checks_10', 'box', 0.7, 'box_0.7'),
    ('checks_10', 'box', 0.6, 'box_0.6'),
    ('checks_10', 'box', 0.5, 'box_0.5'),
    ('checks_5', 'box', 1.2, 'box_1.2'),
    ('checks_5', 'point', 1.2, 'box_1.2'),
    ('checks_5', 'box', 1.4, 'box_1.4'),
    ('checks_5', 'point', 1.4, 'box_1.4'),
    ('checks_5', 'box', 1.6, 'box_1.6'),
    ('checks_5', 'point', 1.6, 'box_1.6'),
    ('checks_5', 'box', 1.8, 'box_1.8'),
    ('checks_5', 'point', 1.8, 'box_1.8'),
    ('checks_5', 'box', 2.0, 'box_2.0'),
    ('checks_5', 'point', 2.0, 'box_2.0'),
    ('checks_10', 'triangle', 0.9, 'triangle_0.9'),
    ('checks_10', 'triangle', 0.8, 'triangle_0.8'),
    ('checks_10', 'triangle', 0.7, 'triangle_0.7'),
    ('checks_10', 'triangle', 0.6, 'triangle_0.6'),
    ('checks_10', 'triangle', 0.5, 'triangle_0.5'),
    ('checks_5', 'triangle', 1.2, 'triangle_1.2'),
    ('checks_5', 'triangle', 1.4, 'triangle_1.4'),
    ('checks_5', 'triangle', 1.6, 'triangle_1.6'),
    ('checks_5', 'triangle', 1.8, 'triangle_1.8'),
    ('checks_5', 'triangle', 2.0, 'triangle_2.0'),
    ('checks_10', 'catrom', 0.9, 'catrom_0.9'),
    ('checks_10', 'catrom', 0.8, 'catrom_0.8'),
    ('checks_10', 'catrom', 0.7, 'catrom_0.7'),
    ('checks_10', 'catrom', 0.6, 'catrom_0.6'),
    ('checks_10', 'catrom', 0.5, 'catrom_0.5'),
    ('checks_5', 'catrom', 1.2, 'catrom_1.2'),
    ('checks_5', 'catrom', 1.4, 'catrom_1.4'),
    ('checks_5', 'catrom', 1.6, 'catrom_1.6'),
    ('checks_5', 'catrom', 1.8, 'catrom_1.8'),
    ('checks_5', 'catrom', 2.0, 'catrom_2.0')])
def test_rescale(image, filter, k, new_name):
    # single channel
    src = read_png(os.path.join(RESOURCES_PATH, image, 'src.png'))
    new = read_png(os.path.join(RESOURCES_PATH, image, new_name + '.png'))
    assert np.allclose(new, clip(rescale(src, k=k, filter=filter)), atol=2)

    # three channel
    src = gray_to_RGB(src)
    new = gray_to_RGB(new)
    assert np.allclose(new, clip(rescale(src, k=k, filter=filter)), atol=2)


@pytest.mark.parametrize("image,k,blur,new_name",[
    ('pixel_5', 300, 0.5, 'blur_0.5'),
    ('pixel_5', 300, 1.0, 'blur_1.0'),
    ('pixel_5', 300, 1.5, 'blur_1.5')])
def test_gaussian_blur(image, k, blur, new_name):
    src = read_png(os.path.join(RESOURCES_PATH, image, 'src.png'))
    new = read_png(os.path.join(RESOURCES_PATH, image, new_name + '.png'))
    assert np.allclose(new, rescale(src, k=k, filter='gaussian', blur=blur),
                       atol=2)


@pytest.mark.parametrize("image,sigma,new_name",[
    ('red_blue_square', 2, 'blur_2'),
    ('red_blue_square', 3, 'blur_3'),
    ('red_blue_square', 5, 'blur_5'),
    ('red_blue_square', 10, 'blur_10'),
    ('red_blue_square', 20, 'blur_20')])
def test_blur(image, sigma, new_name):
    src = read_png(os.path.join(RESOURCES_PATH, image, 'src.png'))
    new = read_png(os.path.join(RESOURCES_PATH, image, new_name + '.png'))
    assert np.allclose(new, blur(src, sigma=sigma), atol=2)


@pytest.mark.parametrize("src,k,new",[
    (np.array([[5,5],[5,5]]), 5, np.array([[5,5],[5,5]])),
    (np.array([[5,5],[5,5]]), 4, np.array([[4,4],[4,4]])),
    (np.array([[-1,5],[5,-1]]), 4, np.array([[0,4],[4,0]]))])
def test_clip(src, k, new):
    assert np.allclose(new, clip(src, k), atol=0)


@pytest.mark.parametrize("src,k,new",[
    (np.array([[5,5],[5,5]]), 4, np.array([[4,4],[4,4]])),
    (np.array([[0,2],[4,6]]), 12, np.array([[0,4],[8,12]])),
    (np.array([[0,2],[4,6]]), 3, np.array([[0,1],[2,3]]))])
def test_normalize(src, k, new):
    assert np.allclose(new, normalize(src, k), atol=0)


@pytest.mark.parametrize("src,k,new",[
    (np.array([[5,5],[5,5]]), 4, np.array([[0,0],[0,0]])),
    (np.array([[0,2],[4,6]]), 12, np.array([[0,2],[4,6]])),
    (np.array([[0,2],[4,6]]), 3, np.array([[0,2],[0,2]]))])
def test_wraparound(src, k, new):
    assert np.allclose(new, wraparound(src, k), atol=0)



import numpy as np
from imageio import imread, imwrite


def read_png(path: str) -> np.ndarray:
    """Read a png file into a NumPy array.

    Args:
        path (str): String file path.

    Returns:
        np.ndarray: NumPy array representing the image.
    """
    image = imread(uri=path, format='png')
    # ignore the transparency channel when reading png files
    if len(image.shape) > 2 and image.shape[2] == 4:
        image = image[:,:,:3]
    return image


def write_png(image: np.ndarray, path: str):
    """Write NumPy array to a png file.

    The NumPy array should have integer values in the range [0, 255].
    Otherwise, this function has undefined behavior.

    Args:
        image (np.ndarray): NumPy array representing image.
        path (str): String file path.
    """
    imwrite(im=image, uri=path, format='png')



import numpy as np
from typing import Callable

# Referenced colorconv.py from scikit-image for more efficient implementation
# of colorspace transformations. Will continue to maintain an independent
# implementation for educational purposes but scikit-image is the standard.

# https://wikipedia.org/wiki/Standard_illuminant
# These values assume the 2 degree point of view
illuminants = \
    {'D50': (96.4212, 100.0, 82.5188),
     'D65': (95.0489, 100.0, 108.8840)}

# http://poynton.ca/PDFs/ColorFAQ.pdf
rgb_to_gray = np.array([0.2125, 0.7154, 0.0721])

# https://wikipedia.org/wiki/CIE_1931_color_space
b_21 = 0.17697
rgb_to_xyz = np.array([[0.49000, 0.31000, 0.20000],
                       [0.17697, 0.81240, 0.01063],
                       [0.00000, 0.01000, 0.99000]]) / b_21
xyz_to_rgb = np.linalg.inv(rgb_to_xyz)

# https://wikipedia.org/wiki/YUV
rgb_to_yuv = np.array([[+0.29900, +0.58700, +0.11400],
                       [-0.14713, -0.28886, +0.43600],
                       [+0.61500, -0.51499, -0.10001]])
yuv_to_rgb = np.linalg.inv(rgb_to_yuv)

# Used to normalize an image in a colorspace to [0,1]
norm = \
    {'RGB': {'scale': (255.0, 255.0, 255.0),
             'shift': (0.0, 0.0, 0.0)},
     'Lab': {'scale': (255.0, 255.0, 255.0),
             'shift': (0.0, -128.0, -128.0)},
     'YUV': {'scale': (1.0, 0.872, 1.23),
             'shift': (0.0, -0.436, -0.615)}}


def RGB_to_gray(image: np.ndarray) -> np.ndarray:
    """Convert an image in CIE RGB space to grayscale.

    For details about the implemented conversion, see
    `FAQs about Color <http://poynton.ca/PDFs/ColorFAQ.pdf>`_.

    Args:
        image (np.ndarray): Image in CIE RGB space.

    Returns:
        np.ndarray: Image in grayscale.
    """
    return image @ rgb_to_gray.T


def gray_to_RGB(image: np.ndarray) -> np.ndarray:
    """Convert an image in grayscale to CIE RGB space.

    Args:
        image (np.ndarray): Image in grayscale.

    Returns:
        np.ndarray: Image in CIE RGB space.
    """
    return np.stack(3 * (image,), axis=-1)


def RGB_to_XYZ(image: np.ndarray) -> np.ndarray:
    """Convert an image in CIE RGB space to XYZ space.

    For details about the implemented conversion, see
    `CIE 1931 color space <https://wikipedia.org/wiki/CIE_1931_color_space>`_.

    Args:
        image (np.ndarray): Image in CIE RGB space.

    Returns:
        np.ndarray: Image in CIE XYZ space.
    """
    return image @ rgb_to_xyz.T


def XYZ_to_RGB(image: np.ndarray) -> np.ndarray:
    """Convert an image in CIE XYZ space to RGB space.

    For details about the implemented conversion, see
    `CIE 1931 color space <https://wikipedia.org/wiki/CIE_1931_color_space>`_.

    Args:
        image (np.ndarray): Image in CIE XYZ space.

    Returns:
        np.ndarray: Image in CIE RGB space.
    """
    return image @ xyz_to_rgb.T


def RGB_to_YUV(image: np.ndarray) -> np.ndarray:
    """Convert an image in CIE RGB space to YUV space.

    For details about the implemented conversion, see
    `YUV <https://wikipedia.org/wiki/YUV>`_.

    Args:
        image (np.ndarray): Image in CIE RGB space.

    Returns:
        np.ndarray: Image in YUV space.
    """
    return normalize(image, 'RGB') @ rgb_to_yuv.T


def YUV_to_RGB(image: np.ndarray) -> np.ndarray:
    """Convert an image in YUV space to CIE RGB space.

    For details about the implemented conversion, see
    `YUV <https://wikipedia.org/wiki/YUV>`_.

    Args:
        image (np.ndarray): Image in YUV space.

    Returns:
        np.ndarray: Image in CIE RGB space.
    """
    return denormalize(image @ yuv_to_rgb.T, 'RGB')


def XYZ_to_Lab(image: np.ndarray, illuminant: str = 'D65') -> np.ndarray:
    """Convert an image in CIE XYZ space to Lab space.

    For details about the implemented conversion, see
    `CIELAB color space <https://wikipedia.org/wiki/CIELAB_color_space>`_.

    Args:
        image (np.ndarray): Image in CIE XYZ space.
        illuminant (str): Standard illuminant {D65, D50}

    Returns:
        np.ndarray: Image in Lab space.
    """
    X_n, Y_n, Z_n = illuminants[illuminant]
    delta = 6 / 29

    def f(t):
        return t**(1/3) if t > delta**3 else (t/(3*delta**2)) + (4/29)

    def to_Lab(x):
        X, Y, Z = x
        L = 116*f(Y/Y_n) - 16
        a = 500*(f(X/X_n) - f(Y/Y_n))
        b = 200*(f(Y/Y_n) - f(Z/Z_n))
        return np.array([L,a,b])

    n,m,k = image.shape
    p = np.reshape(image, (n*m,3)).astype(float)
    p = np.apply_along_axis(to_Lab, 1, p)
    return np.reshape(p, (n,m,k))


def Lab_to_XYZ(image: np.ndarray, illuminant: str = 'D65') -> np.ndarray:
    """Convert an image in Lab space to CIE XYZ space.

    For details about the implemented conversion, see
    `CIELAB color space <https://wikipedia.org/wiki/CIELAB_color_space>`_.

    Args:
        image (np.ndarray): Image in Lab space.
        illuminant (str): Standard illuminant {D65, D50}

    Returns:
        np.ndarray: Image in CIE XYZ space.
    """
    X_n, Y_n, Z_n = illuminants[illuminant]
    delta = 6 / 29

    def f_inv(t):
        return t**3 if t > delta else 3*delta**2*(t-(4/29))

    def to_XYZ(x):
        L, a, b = x
        X = X_n*f_inv(((L + 16)/116) + a/500)
        Y = Y_n*f_inv((L + 16)/116)
        Z = Z_n*f_inv(((L + 16)/116) - b/200)
        return np.array([X,Y,Z])

    n,m,k = image.shape
    p = np.reshape(image, (n*m,3)).astype(float)
    p = np.apply_along_axis(to_XYZ, 1, p)
    return np.reshape(p, (n,m,k))


def RGB_to_Lab(image: np.ndarray, illuminant: str = 'D65') -> np.ndarray:
    """Convert an image in CIE RGB space to Lab space.

    For details about the implemented conversion, see
    `CIE 1931 color space <https://wikipedia.org/wiki/CIE_1931_color_space>`_
    and
    `CIELAB color space <https://wikipedia.org/wiki/CIELAB_color_space>`_.

    Args:
        image (np.ndarray): Image in CIE RGB space.
        illuminant (str): Standard illuminant {D65, D50}

    Returns:
        np.ndarray: Image in Lab space.
    """
    return XYZ_to_Lab(RGB_to_XYZ(image), illuminant)


def Lab_to_RGB(image: np.ndarray, illuminant: str = 'D65') -> np.ndarray:
    """Convert an image in Lab space to CIE RGB space.

    For details about the implemented conversion, see
    `CIE 1931 color space <https://wikipedia.org/wiki/CIE_1931_color_space>`_
    and
    `CIELAB color space <https://wikipedia.org/wiki/CIELAB_color_space>`_.

    Args:
        image (np.ndarray): Image in Lab space.
        illuminant (str): Standard illuminant {D65, D50}

    Returns:
        np.ndarray: Image in CIE RGB space.
    """
    return XYZ_to_RGB(Lab_to_XYZ(image, illuminant))


def normalize(image: np.ndarray, color_space: str) -> np.ndarray:
    """Normalize the image in the given color space.

    Args:
        image (np.ndarray): Image in the given color space.
        color_space (str): Color space {RGB, Lab, YUV}.

    Returns:
        np.ndarray: Normalized image with values in [0,1].
    """
    scale = norm[color_space]['scale']
    shift = norm[color_space]['shift']
    shift_mat = np.ones(image.shape) @ np.diag(shift)
    normalized_image = (image - shift_mat) @ np.diag(1 / np.array(scale))
    return normalized_image.clip(0,1)


def denormalize(image: np.ndarray, color_space: str) -> np.ndarray:
    """Denormalize the image in the given color space.

    Args:
        image (np.ndarray): Normalized image in the given color space.
        color_space (str): Color space {RGB, Lab, YUV}.

    Returns:
        np.ndarray: Denormalized image in the given color space.
    """
    scale = norm[color_space]['scale']
    shift = norm[color_space]['shift']
    shift_mat = np.ones(image.shape) @ np.diag(shift)
    return (image @ np.diag(scale)) + shift_mat


def apply_to_channels(image: np.ndarray,
                      f_1: Callable,
                      f_2: Callable,
                      f_3: Callable) -> np.ndarray:
    """Return the image with the functions applied to each channel.

    Args:
        image (np.ndarray): Image (recommended to be normalized).
        f_1 (Callable): Function to apply to the first channel.
        f_2 (Callable): Function to apply to the second channel.
        f_3 (Callable): Function to apply to the third channel.

    Returns:
        np.ndarray: Pixel matrix with functions applied to each channel.
    """
    n,m,k = image.shape
    p = np.reshape(image, (n*m,3)).astype(float)
    p[:,0] = f_1(p[:,0])
    p[:,1] = f_2(p[:,1])
    p[:,2] = f_3(p[:,2])
    return np.reshape(p, (n,m,k))



import pytest
import numpy as np
from dmtools.colorspace import \
    gray_to_RGB, RGB_to_gray, RGB_to_XYZ, XYZ_to_RGB, RGB_to_YUV, YUV_to_RGB, \
    RGB_to_Lab, Lab_to_RGB

# -----------
# TEST IMAGES
# -----------

# single white pixel (single channel)
ONE_CHANNEL_WHITE_PIXEL = np.array([[255]])

# single white pixel (three channels)
THREE_CHANNEL_WHITE_PIXEL = np.array([[[255, 255, 255]]])

# single color pixel
COLOR_PIXEL = np.array([[[34, 76, 230]]])


@pytest.mark.parametrize("f,source,new",[
    (gray_to_RGB, ONE_CHANNEL_WHITE_PIXEL, THREE_CHANNEL_WHITE_PIXEL),
    (RGB_to_gray, THREE_CHANNEL_WHITE_PIXEL, ONE_CHANNEL_WHITE_PIXEL)])
def test_colorspace_transformation(f, source, new):
    assert np.allclose(new, f(source), atol=1e-6)


@pytest.mark.parametrize("f,f_inv,image",[
    (gray_to_RGB, RGB_to_gray, ONE_CHANNEL_WHITE_PIXEL),
    (RGB_to_XYZ, XYZ_to_RGB, COLOR_PIXEL),
    (RGB_to_XYZ, XYZ_to_RGB, THREE_CHANNEL_WHITE_PIXEL),
    (RGB_to_YUV, YUV_to_RGB, COLOR_PIXEL),
    (RGB_to_YUV, YUV_to_RGB, THREE_CHANNEL_WHITE_PIXEL),
    (RGB_to_Lab, Lab_to_RGB, COLOR_PIXEL),
    (RGB_to_Lab, Lab_to_RGB, THREE_CHANNEL_WHITE_PIXEL)])
def test_colorspace_inverse(f, f_inv, image):
    assert np.allclose(image, f_inv(f(image)), atol=1e-6)



import numpy as np
from scipy.io import wavfile

# https://en.wikipedia.org/wiki/44,100_Hz
SAMPLE_RATE = 44100


class WAV:
    """An object representing a WAV audio file.

    For more information about the audio file format, see
    `WAV <https://en.wikipedia.org/wiki/WAV>`_
    """
    def __init__(self, r: np.ndarray, l: np.ndarray,
                 sample_rate: int = SAMPLE_RATE):
        """Initialize a WAV sound.

        Args:
            r (np.ndarray): NumPy array of samples of the right channel.
            l (np.ndarray): NumPy array of samples of the left channel.
            sample_rate (int): Sample rate. Defaults to SAMPLE_RATE.
        """
        self.r = r
        self.l = l
        self.sample_rate = sample_rate

    def to_wav(self, path):
        """Write object to a WAV audio file (wav)

        Args:
            path (str): String file path.
        """
        wavfile.write(path, self.sample_rate, np.array([self.r, self.l]).T)


def wave(f: float, a: float, t: float) -> np.ndarray:
    """Generate the samples of a sound wave.

    Args:
        f (float): Frequency of the sound wave.
        a (float): Amplitude of the sound wave.
        t (float): Duration (seconds) of the sound wave.

    Returns:
        np.ndarray: NumPy array with sample points of wave.
    """
    sample_points = np.linspace(0, t*(2*np.pi), int(t*SAMPLE_RATE))
    return a*np.sin(sample_points*f)


def wave_sequence(frequencies: np.ndarray, t) -> WAV:
    """Return a Wav sound which iterates through the given frequencies.

    Args:
        frequencies (np.ndarray): frequencies to iterate through.
        t ([type]): duration of iteration.

    Returns:
        WAV: Wav file.
    """
    d = t / len(frequencies)
    w = np.array([list(wave(f,1,d)) for f in frequencies]).flatten()
    return WAV(r=w, l=w)



import os
import imageio
import numpy as np
from .transform import rescale
from typing import List
from ._log import _log_msg
import logging


class Netpbm:
    """An object representing a Netpbm image.

    Netpbm is a package of graphics programs and a programming library. These
    programs work with a set of graphics formats called the "netpbm" formats.
    Each format is identified by a "magic number" which is denoted as :code:`P`
    followed by the number identifier. This class works with the following
    formats.

    - `pbm`_: Pixels are black or white (:code:`P1` and :code:`P4`).
    - `pgm`_: Pixels are shades of gray (:code:`P2` and :code:`P5`).
    - `ppm`_: Pixels are in full color (:code:`P3` and :code:`P6`).

    Each of the formats has two "magic numbers" associated with it. The lower
    number corresponds to the ASCII (plain) format while the higher number
    corresponds to the binary (raw) format. This class can handle reading both
    the plain and raw formats though it can only export Netpbm images in the
    plain formats (:code:`P1`, :code:`P2`, and :code:`P3`).

    The plain formats for all three of pbm, pgm, and ppm are quite similar.
    Here is an example pgm format.

    .. code-block:: text

        P2
        5 3
        4
        1 1 0 1 0
        2 0 3 0 1
        2 2 3 1 0

    The first row of the file contains the "magic number". In this example, the
    file is a grayscale pgm image. The second row gives the file
    dimensions (width by height) separated by whitespace. The third row gives
    the maximum gray/color value. In this case, it is the maximum gray value
    since this is a grayscale pgm image. Essentially, this number encodes how
    many different gradients there are in the image. Lastly, the remaining
    lines of the file encode the actual pixels of the image. In a pbm image,
    the third line is not needed since pixels have binary (black or white)
    values. In a ppm full-color image, each pixels has three values represeting
    it--the values of the red, green, and blue channels.

    This descriptions serves as a brief overview of the Netpbm formats with the
    relevant knowledge for using this class. For more information about Netpbm,
    see the `Netpbm Home Page`_.

    .. _pbm: http://netpbm.sourceforge.net/doc/pbm.html
    .. _pgm: http://netpbm.sourceforge.net/doc/pgm.html
    .. _ppm: http://netpbm.sourceforge.net/doc/ppm.html
    .. _Netpbm Home Page: http://netpbm.sourceforge.net

    """
    extension_to_magic_number = {"pbm": 1, "pgm": 2, "ppm": 3}
    magic_number_to_extension = {1: "pbm", 2: "pgm", 3: "ppm"}

    def __init__(self, P: int, k: int, M: np.ndarray):
        """Initialize a Netpbm image.

        Args:
            P (int): Magic number of the Netpbm image.
            k (int): Maximum gray/color value
            M (np.ndarray): A NumPy array representing the image pixels.
        """
        self.P = P
        self.h, self.w, *_ = M.shape
        self.k = k
        self.M = M

    def __copy__(self):
        return Netpbm(P=self.P, k=self.k, M=self.M)

    def set_max_color_value(self, k: int):
        """Set the maximum gray/color value of this Netpbm image.

        Args:
            k (int): Maximum gray/color value.
        """
        if k == 1:
            self.M = ((self.M / self.k) > 0.5).astype(int)
            if self.P == 2:
                self.P == 1
        else:
            step = int(self.k / k)
            self.M = np.floor_divide(self.M, step)

    def rescale(self, k: int):
        """Rescale the image by the desired scaling factor.

        Uses Nearest-neighbor interpolation as the image scaling algorithm.
        Read more about image scaling algorithms at
        `Image scaling <https://wikipedia.org/wiki/Image_scaling>`_.

        Args:
            k (int): Scale factor
        """
        # Note that order=0 is equivalent to the nearest neighbor algorithm
        self.M = rescale(self.M, k, filter='point')
        self.h, self.w, *_ = self.M.shape

    def to_netpbm(self, path: str, comment: List[str] = []):
        """Write object to a Netpbm file (pbm, pgm, ppm).

        Uses the ASCII (plain) magic numbers.

        Args:
            path (str): String file path.
            comment (str): List of comment lines to include in the file.
        """
        with open(path, "w") as f:
            f.write('P%d\n' % self.P)
            for line in comment:
                f.write('# %s\n' % line)
            f.write("%s %s\n" % (self.w, self.h))
            if self.P != 1:
                f.write("%s\n" % (self.k))
            if self.P == 3:
                M = self.M.reshape(self.h, self.w * 3)
            else:
                M = self.M
            lines = M.clip(0, self.k).astype(int).astype(str).tolist()
            f.write('\n'.join([' '.join(line) for line in lines]))
            f.write('\n')
        logging.info(_log_msg(path, os.stat(path).st_size))

    def to_png(self, path: str):
        """Write object to a png file.

        Args:
            path (str): String file path.
        """
        M = self.M
        if self.P == 1:
            M = np.where(M == 1, 0, 1)
        M = np.array(M * (255 / self.k), dtype=np.uint8)
        imageio.imwrite(path, M)
        logging.info(_log_msg(path, os.stat(path).st_size))


def _parse_ascii_netpbm(f: List[str]) -> Netpbm:
    # adapted from code by Dan Torop
    vals = [v for line in f for v in line.split('#')[0].split()]
    P = int(vals[0][1])
    if P == 1:
        w, h, *vals = [int(v) for v in vals[1:]]
        k = 1
    else:
        w, h, k, *vals = [int(v) for v in vals[1:]]
    if P == 3:
        M = np.array(vals).reshape(h, w, 3)
    else:
        M = np.array(vals).reshape(h, w)
    return Netpbm(P=P, k=k, M=M)


# TODO: make the file reading code more robust
def _parse_binary_netpbm(path: str) -> Netpbm:
    # adapted from https://www.stackvidhya.com/python-read-binary-file/
    with open(path, "rb") as f:
        P = int(f.readline().decode()[1])
        # change to corresponding ASCII magic number
        P = int(P / 2)
        w = int(f.readline().decode()[:-1])
        h = int(f.readline().decode()[:-1])
        if P == 1:
            k = 1
        else:
            k = int(f.readline().decode()[:-1])
        dtype = np.dtype('B')
        M = np.fromfile(f, dtype)
        if P == 3:
            M = M.reshape(h, w, 3)
        else:
            M = M.reshape(h, w)
    return Netpbm(P=P, k=k, M=M)


def read_netpbm(path: str) -> Netpbm:
    """Read Netpbm file (pbm, pgm, ppm) into Netpbm.

    Args:
        path (str): String file path.

    Returns:
        Netpbm: A Netpbm image
    """
    with open(path, "rb") as f:
        magic_number = f.read(2).decode()
    if int(magic_number[1]) <= 3:
        # P1, P2, P3 are the ASCII (plain) formats
        with open(path) as f:
            return _parse_ascii_netpbm(f)
    else:
        # P4, P5, P6 are the binary (raw) formats
        return _parse_binary_netpbm(path)



import numpy as np
from typing import List


def image_grid(images: List[np.ndarray], w: int, h: int, b: int,
               color: int = "white", k: int = 255) -> np.ndarray:
    """Create a w * h grid of images with a border of width b.

    Args:
        images (List[np.ndarray]): images (of same dimension) for grid.
        w (int): number of images in each row of the grid.
        h (int): number of images in each column of the grid.
        b (int): width of the border/margin.
        color (int): color of border {'white', 'black'} (defaults to white).
        k (int): white point (defaults to 255).

    Returns:
        np.ndarray: grid layout of the images.
    """
    n,m = images[0].shape
    c = {'white': k, 'black': 0}[color]
    h_border = c*np.ones((b, w*m + (w+1)*b))
    v_border = c*np.ones((n, b))
    grid_layout = h_border
    p = 0
    for i in range(h):
        row = v_border
        for j in range(w):
            row = np.hstack((row, images[p]))
            row = np.hstack((row, v_border))
            p += 1
        grid_layout = np.vstack((grid_layout, row))
        grid_layout = np.vstack((grid_layout, h_border))
    return grid_layout.astype(int)


def border(image: np.ndarray, b: int,
           color: int = "white", k: int = 255) -> np.ndarray:
    """Add a border of width b to the image.

    Args:
        image (Netpbm): Netpbm image to add a border to
        b (int): width of the border/margin.
        color (int): color of border {'white', 'black'} (defaults to white).
        k (int): white point (defaults to 255).

    Returns:
        np.ndarray: Image with border added.
    """
    return image_grid([image], w=1, h=1, b=b, color=color, k=k)



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="dmtools",
    version="0.3.0",
    author="Henry Robbins",
    author_email="hwr26@cornell.edu",
    description="A Python package providing low-level tools for working with "
                "digital media programmatically.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/henryrobbins/dmtools.git",
    packages=find_packages(),
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'numpy>=1.19',
        'imageio>=2.9',
        'imageio-ffmpeg>=0.4.5',
        'scipy>=1.6',
        'typing>=3.7'
    ],
    extras_require= {
        "dev": ['pytest>=5',
                'mock>=3',
                'coverage>=4.5',
                'tox>=3']
    },
    python_requires='>=3.5',
)


import imageio
import numpy as np
from math import ceil
from typing import List
import logging
from . import sound
from ._log import _log_msg
import os


def clip(path: str, start: int = 0, end: int = -1) -> List[np.ndarray]:
    """Return a list of images in the given directory.

    Images are ordered according to their name. Hence, the following naming
    convention is recommend.

    name0000.png, name0001.png, ...

    Args:
        path (str): String directory path.
        start (int, optional): Starting frame. Defaults to 0.
        end (int, optional): Ending frame. Defaults to -1.

    Returns:
        List[np.ndarray]: List of NumPy arrays representing images.
    """
    def listdir_nohidden(path):
        for f in os.listdir(path):
            if not f.startswith('.'):
                yield f

    files = sorted(listdir_nohidden(path))
    files = files[start:end]
    paths = ["%s/%s" % (path, f) for f in files]
    frames = [imageio.imread(path) for path in paths]
    return frames


def _pad_to_16(M: np.ndarray) -> np.ndarray:
    # TODO: Get a better understanding why image demensions need to be
    # multiplies of 16. It appears this requirement is no longer from ffmpeg.
    # Adapted from code by: https://stackoverflow.com/users/9698684/yatu
    m,n,*_ = M.shape
    y_pad = (ceil(m/16)*16-m)
    y_pad_split = (y_pad // 2, y_pad // 2 + y_pad % 2)
    x_pad = (ceil(n/16)*16-n)
    x_pad_split = (x_pad // 2, x_pad // 2 + x_pad % 2)
    if len(M.shape) == 3:
        return np.pad(M, (y_pad_split, x_pad_split, (0, 0)))
    else:
        return np.pad(M, (y_pad_split, x_pad_split))


def to_mp4(frames: List[np.ndarray], path: str, fps: int, s: int = 1,
           audio: sound.WAV = None):
    """Write an animation as a .mp4 file using ffmpeg through imageio.mp4

    Args:
        frames (List[np.ndarray]): List of frames in the animation.
        audio (sound.WAV): Audio for the animation (None if no audio).
        path (str): String file path.
        fps (int): Frames per second.
        s (int, optional): Multiplier for scaling. Defaults to 1.
    """
    frames = [f.astype(np.uint8).clip(0, 255) for f in frames]
    frames = [_pad_to_16(f) for f in frames]
    imageio.mimwrite(uri="tmp.mp4" if audio is not None else path,
                     ims=frames,
                     format='FFMPEG',
                     fps=fps,
                     output_params=["-vf", "scale=iw*%d:ih*%d" % (s, s),
                                    "-sws_flags", "neighbor"])
    if audio is not None:
        audio.to_wav("tmp.wav")
        os.system("ffmpeg -i %s -i %s -c:v copy -c:a aac -y %s"
                  % ("tmp.mp4", "tmp.wav", path))
        os.system("rm tmp.mp4")
        os.system("rm tmp.wav")
    name = path.split('/')[-1]
    logging.info(_log_msg(name, os.stat(path).st_size))



import pytest
import numpy as np
from dmtools.arrange import image_grid, border

# -----------
# TEST IMAGES
# -----------

# single white pixel
WHITE_PIXEL = np.array([[255]])

# single white pixel with 2 pixel black border
WHITE_PIXEL_BLACK_BORDER = \
    np.array([[  0,   0,   0,   0,   0],
              [  0,   0,   0,   0,   0],
              [  0,   0, 255,   0,   0],
              [  0,   0,   0,   0,   0],
              [  0,   0,   0,   0,   0]])

# four white pixels in 2x2 grid with 1 pixel black border
FOUR_WHITE_PIXEL_GRID = \
    np.array([[  0,   0,   0,   0,   0],
              [  0, 255,   0, 255,   0],
              [  0,   0,   0,   0,   0],
              [  0, 255,   0, 255,   0],
              [  0,   0,   0,   0,   0]])


@pytest.mark.parametrize("images,w,h,b,color,result",[
    ([WHITE_PIXEL], 1, 1, 2, 'black', WHITE_PIXEL_BLACK_BORDER),
    ([WHITE_PIXEL]*4, 2, 2, 1, 'black', FOUR_WHITE_PIXEL_GRID)])
def test_image_grid(images, w, h, b, color, result):
    assert np.array_equal(result, image_grid(images, w, h, b, color))


@pytest.mark.parametrize("image,b,color,result",[
    (WHITE_PIXEL, 2, 'black', WHITE_PIXEL_BLACK_BORDER)])
def test_border(image, b, color, result):
    assert np.array_equal(result, border(image, b, color))
